@@ -1,20 +1,11 @@
 # frozen_string_literal: true
 
-require 'puma/rack/builder'
-require 'puma/plugin'
-require 'puma/const'
+require_relative 'plugin'
+require_relative 'const'
+require_relative 'dsl'
+require_relative 'events'
 
 module Puma
-
-  module ConfigDefault
-    DefaultRackup = "config.ru"
-
-    DefaultTCPHost = "0.0.0.0"
-    DefaultTCPPort = 9292
-    DefaultWorkerTimeout = 60
-    DefaultWorkerShutdownTimeout = 30
-  end
-
   # A class used for storing "leveled" configuration options.
   #
   # In this class any "user" specified options take precedence over any
@@ -122,7 +113,7 @@ module Puma
   #     config = Configuration.new({}) do |user_config, file_config, default_config|
   #       user_config.port 3003
   #     end
-  #     config.load
+  #     config.clamp
   #     puts config.options[:port]
   #     # => 3003
   #
@@ -135,27 +126,83 @@ module Puma
   # is done because an environment variable may have been modified while loading
   # configuration files.
   class Configuration
-    include ConfigDefault
+    class NotLoadedError < StandardError; end
+    class NotClampedError < StandardError; end
 
-    def initialize(user_options={}, default_options = {}, &block)
-      default_options = self.puma_default_options.merge(default_options)
+    DEFAULTS = {
+      auto_trim_time: 30,
+      binds: ['tcp://0.0.0.0:9292'.freeze],
+      fiber_per_request: !!ENV.fetch("PUMA_FIBER_PER_REQUEST", false),
+      debug: false,
+      enable_keep_alives: true,
+      early_hints: nil,
+      environment: 'development'.freeze,
+      # Number of seconds to wait until we get the first data for the request.
+      first_data_timeout: 30,
+      # Number of seconds to wait until the next request before shutting down.
+      idle_timeout: nil,
+      io_selector_backend: :auto,
+      log_requests: false,
+      logger: STDOUT,
+      # Limits how many requests a keep alive connection can make.
+      # The connection will be closed after it reaches `max_keep_alive`
+      # requests.
+      max_keep_alive: 999,
+      max_threads: Puma.mri? ? 5 : 16,
+      min_threads: 0,
+      mode: :http,
+      mutate_stdout_and_stderr_to_sync_on_write: true,
+      out_of_band: [],
+      # Number of seconds for another request within a persistent session.
+      persistent_timeout: ENV.fetch('PUMA_PERSISTENT_TIMEOUT', 65),
+      queue_requests: true,
+      rackup: 'config.ru'.freeze,
+      raise_exception_on_sigterm: true,
+      reaping_time: 1,
+      remote_address: :socket,
+      silence_single_worker_warning: false,
+      silence_fork_callback_warning: false,
+      tag: File.basename(Dir.getwd),
+      tcp_host: '0.0.0.0'.freeze,
+      tcp_port: 9292,
+      wait_for_less_busy_worker: 0.005,
+      worker_boot_timeout: 60,
+      worker_check_interval: 5,
+      worker_culling_strategy: :youngest,
+      worker_shutdown_timeout: 30,
+      worker_timeout: 60,
+      workers: 0,
+      http_content_length_limit: nil
+    }
 
-      @options     = UserFileDefaultOptions.new(user_options, default_options)
+    def initialize(user_options={}, default_options = {}, env = ENV, &block)
+      default_options = self.puma_default_options(env).merge(default_options)
+
+      @_options    = UserFileDefaultOptions.new(user_options, default_options)
       @plugins     = PluginLoader.new
-      @user_dsl    = DSL.new(@options.user_options, self)
-      @file_dsl    = DSL.new(@options.file_options, self)
-      @default_dsl = DSL.new(@options.default_options, self)
+      @events      = @_options[:events] || Events.new
+      @hooks       = {}
+      @user_dsl    = DSL.new(@_options.user_options, self)
+      @file_dsl    = DSL.new(@_options.file_options, self)
+      @default_dsl = DSL.new(@_options.default_options, self)
 
-      if !@options[:prune_bundler]
-        default_options[:preload_app] = (@options[:workers] > 1) && Puma.forkable?
-      end
+      @puma_bundler_pruned = env.key? 'PUMA_BUNDLER_PRUNED'
 
       if block
         configure(&block)
       end
+
+      @loaded = false
+      @clamped = false
     end
 
-    attr_reader :options, :plugins
+    attr_reader :plugins, :events, :hooks
+
+    def options
+      raise NotClampedError, "ensure clamp is called before accessing options" unless @clamped
+
+      @_options
+    end
 
     def configure
       yield @user_dsl, @file_dsl, @default_dsl
@@ -168,7 +215,7 @@ module Puma
     def initialize_copy(other)
       @conf        = nil
       @cli_options = nil
-      @options     = @options.dup
+      @_options     = @_options.dup
     end
 
     def flatten
@@ -176,55 +223,49 @@ module Puma
     end
 
     def flatten!
-      @options = @options.flatten
+      @_options = @_options.flatten
       self
     end
 
-    # @version 5.0.0
-    def default_max_threads
-      Puma.mri? ? 5 : 16
+    def puma_default_options(env = ENV)
+      defaults = DEFAULTS.dup
+      puma_options_from_env(env).each { |k,v| defaults[k] = v if v }
+      defaults
     end
 
-    def puma_default_options
+    def puma_options_from_env(env = ENV)
+      min = env['PUMA_MIN_THREADS'] || env['MIN_THREADS']
+      max = env['PUMA_MAX_THREADS'] || env['MAX_THREADS']
+      workers = if env['WEB_CONCURRENCY'] == 'auto'
+        require_processor_counter
+        ::Concurrent.available_processor_count
+      else
+        env['WEB_CONCURRENCY']
+      end
+
       {
-        :min_threads => Integer(ENV['PUMA_MIN_THREADS'] || ENV['MIN_THREADS'] || 0),
-        :max_threads => Integer(ENV['PUMA_MAX_THREADS'] || ENV['MAX_THREADS'] || default_max_threads),
-        :log_requests => false,
-        :debug => false,
-        :binds => ["tcp://#{DefaultTCPHost}:#{DefaultTCPPort}"],
-        :workers => Integer(ENV['WEB_CONCURRENCY'] || 0),
-        :silence_single_worker_warning => false,
-        :mode => :http,
-        :worker_timeout => DefaultWorkerTimeout,
-        :worker_boot_timeout => DefaultWorkerTimeout,
-        :worker_shutdown_timeout => DefaultWorkerShutdownTimeout,
-        :remote_address => :socket,
-        :tag => method(:infer_tag),
-        :environment => -> { ENV['APP_ENV'] || ENV['RACK_ENV'] || ENV['RAILS_ENV'] || 'development' },
-        :rackup => DefaultRackup,
-        :logger => STDOUT,
-        :persistent_timeout => Const::PERSISTENT_TIMEOUT,
-        :first_data_timeout => Const::FIRST_DATA_TIMEOUT,
-        :raise_exception_on_sigterm => true,
-        :max_fast_inline => Const::MAX_FAST_INLINE,
-        :io_selector_backend => :auto,
-        :mutate_stdout_and_stderr_to_sync_on_write => true,
+        min_threads: min && min != "" && Integer(min),
+        max_threads: max && max != "" && Integer(max),
+        workers: workers && workers != "" && Integer(workers),
+        environment: env['APP_ENV'] || env['RACK_ENV'] || env['RAILS_ENV'],
       }
     end
 
     def load
+      @loaded = true
       config_files.each { |config_file| @file_dsl._load_from(config_file) }
-
-      @options
+      @_options
     end
 
     def config_files
-      files = @options.all_of(:config_files)
+      raise NotLoadedError, "ensure load is called before accessing config_files" unless @loaded
+
+      files = @_options.all_of(:config_files)
 
       return [] if files == ['-']
       return files if files.any?
 
-      first_default_file = %W(config/puma/#{environment_str}.rb config/puma.rb).find do |f|
+      first_default_file = %W(config/puma/#{@_options[:environment]}.rb config/puma.rb).find do |f|
         File.exist?(f)
       end
 
@@ -232,9 +273,16 @@ module Puma
     end
 
     # Call once all configuration (included from rackup files)
-    # is loaded to flesh out any defaults
+    # is loaded to finalize defaults and lock in the configuration.
+    #
+    # This also calls load if it hasn't been called yet.
     def clamp
-      @options.finalize_values
+      load unless @loaded
+      set_conditional_default_options
+      @_options.finalize_values
+      @clamped = true
+      warn_hooks
+      options
     end
 
     # Injects the Configuration object into the env
@@ -253,11 +301,11 @@ module Puma
     # Indicate if there is a properly configured app
     #
     def app_configured?
-      @options[:app] || File.exist?(rackup)
+      options[:app] || File.exist?(rackup)
     end
 
     def rackup
-      @options[:rackup]
+      options[:rackup]
     end
 
     # Load the specified rackup file, pull options from
@@ -266,9 +314,9 @@ module Puma
     def app
       found = options[:app] || load_rackup
 
-      if @options[:log_requests]
-        require 'puma/commonlogger'
-        logger = @options[:logger]
+      if options[:log_requests]
+        require_relative 'commonlogger'
+        logger = options[:custom_logger] ? options[:custom_logger] : options[:logger]
         found = CommonLogger.new(found, logger)
       end
 
@@ -277,30 +325,37 @@ module Puma
 
     # Return which environment we're running in
     def environment
-      @options[:environment]
-    end
-
-    def environment_str
-      environment.respond_to?(:call) ? environment.call : environment
+      options[:environment]
     end
 
     def load_plugin(name)
       @plugins.create name
     end
 
-    def run_hooks(key, arg, events)
-      @options.all_of(key).each do |b|
+    # @param key [:Symbol] hook to run
+    # @param arg [Launcher, Int] `:before_restart` passes Launcher
+    #
+    def run_hooks(key, arg, log_writer, hook_data = nil)
+      log_writer.debug "Running #{key} hooks"
+
+      options.all_of(key).each do |hook_options|
         begin
-          b.call arg
+          block = hook_options[:block]
+          if id = hook_options[:id]
+            hook_data[id] ||= Hash.new
+            block.call arg, hook_data[id]
+          else
+            block.call arg
+          end
         rescue => e
-          events.log "WARNING hook #{key} failed with exception (#{e.class}) #{e.message}"
-          events.debug e.backtrace.join("\n")
+          log_writer.log "WARNING hook #{key} failed with exception (#{e.class}) #{e.message}"
+          log_writer.debug e.backtrace.join("\n")
         end
       end
     end
 
     def final_options
-      @options.final_options
+      options.final_options
     end
 
     def self.temp_path
@@ -310,10 +365,22 @@ module Puma
       "#{Dir.tmpdir}/puma-status-#{t}-#{$$}"
     end
 
+    def self.random_token
+      require 'securerandom' unless defined?(SecureRandom)
+
+      SecureRandom.hex(16)
+    end
+
     private
 
-    def infer_tag
-      File.basename(Dir.getwd)
+    def require_processor_counter
+      require 'concurrent/utility/processor_counter'
+    rescue LoadError
+      warn <<~MESSAGE
+        WEB_CONCURRENCY=auto requires the "concurrent-ruby" gem to be installed.
+        Please add "concurrent-ruby" to your Gemfile.
+      MESSAGE
+      raise
     end
 
     # Load and use the normal Rack builder if we can, otherwise
@@ -321,7 +388,7 @@ module Puma
     def rack_builder
       # Load bundler now if we can so that we can pickup rack from
       # a Gemfile
-      if ENV.key? 'PUMA_BUNDLER_PRUNED'
+      if @puma_bundler_pruned
         begin
           require 'bundler/setup'
         rescue LoadError
@@ -331,11 +398,10 @@ module Puma
       begin
         require 'rack'
         require 'rack/builder'
+        ::Rack::Builder
       rescue LoadError
-        # ok, use builtin version
-        return Puma::Rack::Builder
-      else
-        return ::Rack::Builder
+        require_relative 'rack/builder'
+        Puma::Rack::Builder
       end
     end
 
@@ -345,24 +411,40 @@ module Puma
       rack_app, rack_options = rack_builder.parse_file(rackup)
       rack_options = rack_options || {}
 
-      @options.file_options.merge!(rack_options)
+      options.file_options.merge!(rack_options)
 
       config_ru_binds = []
       rack_options.each do |k, v|
         config_ru_binds << v if k.to_s.start_with?("bind")
       end
 
-      @options.file_options[:binds] = config_ru_binds unless config_ru_binds.empty?
+      options.file_options[:binds] = config_ru_binds unless config_ru_binds.empty?
 
       rack_app
     end
 
-    def self.random_token
-      require 'securerandom' unless defined?(SecureRandom)
+    def set_conditional_default_options
+      @_options.default_options[:preload_app] = !@_options[:prune_bundler] &&
+        (@_options[:workers] > 1) && Puma.forkable?
+    end
 
-      SecureRandom.hex(16)
+    def warn_hooks
+      return if options[:workers] > 0
+      return if options[:silence_fork_callback_warning]
+
+      log_writer = LogWriter.stdio
+      @hooks.each_key do |hook|
+        options.all_of(hook).each do |hook_options|
+          next unless hook_options[:cluster_only]
+
+          log_writer.log(<<~MSG.tr("\n", " "))
+            Warning: The code in the `#{hook}` block will not execute
+            in the current Puma configuration. The `#{hook}` block only
+            executes in Puma's cluster mode. To fix this, either remove the
+            `#{hook}` call or increase Puma's worker count above zero.
+          MSG
+        end
+      end
     end
   end
 end
-
-require 'puma/dsl'
